/////////////////////////////////////////////////////////////////////////////
//// INCLUDE FILES

#define		DEVICE_MAIN
#include	"SNTRANS.h"
#undef		DEVICE_MAIN

// Copyright And Configuration Management ----------------------------------
//
//        NDIS 3.X Network Protocol (Transport) Example - SNTRANS.c
//                    For Microsoft Windows And VtoolsD
//
//     Copyright (c) 1995-1997, Printing Communications Associates, Inc.
//
//                             Thomas F. Divine
//                           4201 Brunswick Court
//                        Smyrna, Georgia 30080 USA
//                              (770) 436-1714
//
// $Workfile$
//  $Logfile$
//   $Author: FRED $
//     $Date: 1995/11/28 08:35:30 $
// $Revision: 1.1 $
//
// $Log: SNTRANS.c $
// Revision 1.1  1995/11/28 08:35:30  FRED
// Initial revision
// 
// End ---------------------------------------------------------------------

// This example is the framework for a NDIS 3.1 PROTOCOL driver. It provides
// a structure on which to build an NDIS driver for a proprietary
// network protocol. To actually implement a driver requires creation of 
// all the routines named in the NDIS_PROTOCOL_CHARACTERISTICS structure 
// passed to NdisRegisterProtocol. Stubs for these routines are found in
// TRANSFCN.C.
//
// Note that an NDIS 3.1 PROTOCOL driver is dynamically-loaded and begins
// execution with the SYS_DYNAMIC_DEVICE_INIT message. The initialization
// work is done in OnSysDynamicDeviceInit(), BindAdapterHandler() and
// OpenAdapterCompleteHandler().
//
// Read the comments in the NETSN31.INF Network Device Installer (NDI)
// information file for clues on how to install SNTRANS.VXD.
//

// -------------------------------------------------------------------
// Device Preliminaries

Declare_Virtual_Device(SNTRANS)		// Create Device Descriptor Block

CHAR VendorString[] = TRANSPORT_NAME;		// For VxD Access

// ------------------------------------------------------------------------
// Declare prototypes for control message handlers

#ifdef WIN40
DefineControlHandler(SYS_DYNAMIC_DEVICE_INIT, OnSysDynamicDeviceInit);
DefineControlHandler(SYS_DYNAMIC_DEVICE_EXIT, OnSysDynamicDeviceExit);
DefineControlHandler(W32_DEVICEIOCONTROL, OnW32Deviceiocontrol);
#else
DefineControlHandler(BEGIN_PM_APP, OnBeginPmApp);
DefineControlHandler(END_PM_APP, OnEndPmApp);
DefineControlHandler(DEVICE_INIT, OnDeviceInit);
DefineControlHandler(SYSTEM_EXIT, OnSystemExit);
#endif


// -------------------------------------------------------------------
// Static data


// -------------------------------------------------------------------
// Control Message Dispatcher

BOOL ControlDispatcher(
	DWORD dwControlMessage,
	DWORD EBX,
	DWORD EDX,
	DWORD ESI,
	DWORD EDI,
	DWORD ECX)
{
	START_CONTROL_DISPATCH

#ifdef WIN40
		ON_SYS_DYNAMIC_DEVICE_INIT(OnSysDynamicDeviceInit);
		ON_SYS_DYNAMIC_DEVICE_EXIT(OnSysDynamicDeviceExit);
		ON_W32_DEVICEIOCONTROL(OnW32Deviceiocontrol);
#else
		ON_BEGIN_PM_APP(OnBeginPmApp);
		ON_END_PM_APP(OnEndPmApp);
		ON_DEVICE_INIT(OnDeviceInit);
		ON_SYSTEM_EXIT(OnSystemExit);
#endif

	END_CONTROL_DISPATCH

	return TRUE;
}


#ifdef WIN40

// ------------------------------------------------------------------------
// Handler for W32_DEVICEIOCONTROL
//
// This function is called when
//
//  (1) The app calls CreateFile
//  (2) The app calls DeviceIOControl
//  (3) The app exits (or calls CloseHandle)
//
DWORD OnW32Deviceiocontrol(PIOCTLPARAMS p)
{
	/* Sanity Check
	--------------- */
	Assert( p );

	if( !p )
	{
		return (DWORD)0xffffffff;
	}

	dprintf("SNTRANS: entered with function=%d\n", p->dioc_IOCtlCode);

// Structure member dioc_IOCtlCode is 

	switch (p->dioc_IOCtlCode)
	{
		case DIOC_OPEN:		// CreateFile
		case DIOC_CLOSEHANDLE: 	// file closed
		return 0;

#ifdef ZNEVER
	case W32IF_PASS_EVENT:

// First, store the ring 0 event handle being passed in

		hEventToNotifyApp = *(HANDLE*)p->dioc_InBuf;

// Set the return information to the addresses of the relevant variables

		((DWORD*)(p->dioc_OutBuf))[0] =	(DWORD)&handleOfLastCreatedOrDestroyedVM;
		((DWORD*)(p->dioc_OutBuf))[1] = (DWORD)&bLastOperation;

// Set the size of the data returned

		*p->dioc_bytesret = 2*sizeof(DWORD);

// Return OK
		return 0;
#endif // ZNEVER

	default:
		return (DWORD)0xffffffff;
	}
}

// -------------------------------------------------------------------
// Handler for Control Message SYS_DYNAMIC_DEVICE_EXIT

BOOL OnSysDynamicDeviceExit()
{
	DEBUGTRACE( "OnSysDynamicDeviceExit Entry..." );

	return TRUE;
}

#else


// -------------------------------------------------------------------
// VXD PM Vendor Specific Device Entry Point
//
// PM applications locate this entry point by passing the VxD's device
// identification string to the Windows Interrupt 2F Get Vendor Specific
// Device Entry Point (168A) service.
//
// Be sure to understand that the term PM means "16-bit protected mode
// application. If you need to call this entry point from Win32s applications
// running under WFW, then you will have to use a mechanism called "universal
// thunks". Good luck!!!

VOID __stdcall PM_Vendor_Api_Handler(
						VMHANDLE hVM,
						PVOID pRefdata,
						PCLIENT_STRUCT pRegs
						)
{
// Handle Vendor API requests from protected mode here.
// Handler must call Simulate_Far_Ret() before returning.

	_clientEAX = 0;		// Success

	Simulate_Far_Ret();
}

BOOL OnBeginPmApp(VMHANDLE hVM, DWORD flags, PPMCB pmcb)
{
	Set_PM_Vendor_Entry(VendorString, PM_Vendor_Api_Handler);

	return TRUE;
}

VOID OnEndPmApp(VMHANDLE hVM, PPMCB pmcb)
{
	End_PM_Vendor_Entry();
}

// -------------------------------------------------------------------
// Handler for Control Message SYSTEM_EXIT

VOID OnSystemExit( VMHANDLE hSysVM )
{
	NDIS_STATUS	nNdisStatus;

	DEBUGTRACE( "OnSystemExit Entry..." );

	/* Unbind The Adapter
	---------------------
	 * Note: This mechanism will need to be modified if the protocol
	 * is bound to multiple adapters...
	 */
	SNTRANSUnbindAdapterHandler(
    	&nNdisStatus,
		(NDIS_HANDLE )&SNTRANSContext.hNdisProtocolHandle,
    	NULL
    	);

	/* Unload The Protocol
	---------------------- */
	SNTRANSUnloadProtocolHandler();
}

#endif


// -------------------------------------------------------------------
// Main entry point for NDIS driver

NTSTATUS NDIS_API DriverEntry(
	IN PDRIVER_OBJECT DriverObject,		// NULL for VxDs
	IN PUNICODE_STRING RegistryPath		// NULL for VxDs
)
{
	NDIS_STATUS		nNdisStatus, nNdisStatus2;
	NDIS_STRING		SNTRANSName = NDIS_STRING_CONST(TRANSPORT_NAME);

	DEBUGTRACE( "DriverEntry Entry..." );

//
// Under NDIS 3.1 PROTOCOL drivers are dynamically loaded and start
// executing by receiving the Sys_Dynamic_Init message. They then
// may call NdisOpenProtocolConfiguration() and NdisReadConfiguration()
// to to get global configuration information from the Registry.
//

//
// Initialize the PROTOCOL characteristics for the call to
// NdisRegisterPrtocol.
//

	SNTRANSContext.ProtocolCharacteristics.MajorNdisVersion = SNTRANS_NDIS_MAJOR_VERSION;
	SNTRANSContext.ProtocolCharacteristics.MinorNdisVersion = SNTRANS_NDIS_MINOR_VERSION;
	SNTRANSContext.ProtocolCharacteristics.Reserved = 0;
	SNTRANSContext.ProtocolCharacteristics.OpenAdapterCompleteHandler = SNTRANSOpenAdapterCompleteHandler;
	SNTRANSContext.ProtocolCharacteristics.CloseAdapterCompleteHandler = SNTRANSCloseAdapterCompleteHandler;
	SNTRANSContext.ProtocolCharacteristics.SendCompleteHandler = SNTRANSSendCompleteHandler;
	SNTRANSContext.ProtocolCharacteristics.TransferDataCompleteHandler = SNTRANSTransferDataCompleteHandler;
	SNTRANSContext.ProtocolCharacteristics.ResetCompleteHandler = SNTRANSResetCompleteHandler;
	SNTRANSContext.ProtocolCharacteristics.RequestCompleteHandler = SNTRANSRequestCompleteHandler;
	SNTRANSContext.ProtocolCharacteristics.ReceiveHandler = SNTRANSReceiveHandler;
	SNTRANSContext.ProtocolCharacteristics.ReceiveCompleteHandler = SNTRANSReceiveCompleteHandler;
	SNTRANSContext.ProtocolCharacteristics.StatusHandler = SNTRANSStatusHandler;
	SNTRANSContext.ProtocolCharacteristics.StatusCompleteHandler = SNTRANSStatusCompleteHandler;

#ifdef WIN40
	SNTRANSContext.ProtocolCharacteristics.BindAdapterHandler = SNTRANSBindAdapterHandler;
	SNTRANSContext.ProtocolCharacteristics.UnbindAdapterHandler = SNTRANSUnbindAdapterHandler;
	SNTRANSContext.ProtocolCharacteristics.UnloadProtocolHandler = SNTRANSUnloadProtocolHandler;
#endif

	SNTRANSContext.ProtocolCharacteristics.Name = SNTRANSName;

// The NdisRegisterProtocol request is called when a PROTOCOL module
// initializes. It provides the NDIS interface with information about the
// PROTOCOL driver, including the addresses of its request handlers.

// The PROTOCOL driver passes in a characteristics table. This table is
// copied by the NdisRegisterProtocol request to its own internal storage.
// Thus, once registered, the PROTOCOL driver cannot alter its handler
// routines.

	DEBUGTRACE( "Calling NdisRegisterProtocol..." );

	NdisRegisterProtocol(
		&nNdisStatus,
		&SNTRANSContext.hNdisProtocolHandle,
		&SNTRANSContext.ProtocolCharacteristics,
		sizeof( NDIS_PROTOCOL_CHARACTERISTICS )
		);

	if( nNdisStatus != NDIS_STATUS_SUCCESS )
	{
		//
		// NdisRegisterProtocol failed.
		//

#ifdef DEBUG
		dprintf( "NdisRegisterProtocol failed with code 0x%X\n", nNdisStatus );
#endif

		return nNdisStatus;
	}

	DEBUGTRACE( "NdisRegisterProtocol succeeded\n" );

#ifndef WIN40


#endif

	DEBUGTRACE( "Protocol Initialization Complete\n" );

	return( NDIS_STATUS_SUCCESS );
}


#ifdef WIN40

// -------------------------------------------------------------------
// Handler for Control Message SYS_DYNAMIC_DEVICE_INIT (NDIS 3.1/Win 95)

BOOL OnSysDynamicDeviceInit()
{
	SETDEBUGLEVEL( DBG_ERROR, DBG_TRACE );

	DEBUGTRACE( "OnSysDynamicDeviceInit Entry..." );

	/* Call The Common NDIS 3.0/3.1 Driver Entry Point
	-------------------------------------------------- */
	return (DriverEntry(NULL, NULL) == NDIS_STATUS_SUCCESS);
}

#endif // WIN40

// -------------------------------------------------------------------
// Remaining code in this module is intialization code

#include INIT_CODE_SEGMENT

#ifndef WIN40

// -------------------------------------------------------------------
// Find the length of the current line from the mapped file data
// pointed to by pBufPtr.
//
// The returned length includes the terminating newline and/or
// carriage return characters, if present.

BOOL
INIGetLineLength(
	IN char	*pBufPtr,					// Start Of Current Line
	IN UINT	nBufSize,					// Size Of Data At pBufPtr
	OUT UINT	*nLineLength
	)
{
	/* Sanity Checks On Arguments
	----------------------------- */
	if( nBufSize == 0 )
	{
		return( FALSE );
	}

	Assert( pBufPtr );
	Assert( nLineLength );

	if( !pBufPtr || !nLineLength )
	{
		return( FALSE );
	}

	*nLineLength = 0;

	/* Walk The Character Buffer Looking For End Of Current Line
	------------------------------------------------------------ */
	while( nBufSize )
	{
		/* Look For Newline
		------------------- */
		if( *pBufPtr == '\n' )
		{
			/* Move To The Next Character
			----------------------------- */
			++pBufPtr;
			--nBufSize;
			*nLineLength += 1;

			/* Found Newline
			---------------- */
			if( nBufSize > 0 )
			{
				/* Look For Carriage Return Following Newline
				--------------------------------------------- */
				if( *pBufPtr == '\r' )
				{
					/* Found Carriage Return - Include It In Current Line
					----------------------------------------------------- */
					++pBufPtr;
					--nBufSize;
					*nLineLength += 1;
				}
			}

			return( TRUE );						// Found EOL
		}

		/* Move To The Next Character
		----------------------------- */
		++pBufPtr;
		--nBufSize;
		*nLineLength += 1;
	}

	if( *nLineLength > 0 )
	{
		/* Found End Of Data Without Terminating newline Or CR
		------------------------------------------------------ */
		return( TRUE );
	}

	return( FALSE );								// EOL Not Found
}


// -------------------------------------------------------------------
// Perform a case-insensitive comparision between the two 'C' strings
// pointed to by s and t.

int INIstrnicmp( char *s, char *t, UINT nLineLength )
{
	/* Sanity Check On Arguments
	---------------------------- */
	Assert( s );
	Assert( t );
	Assert( nLineLength > 0 );

	if( !s || !t || nLineLength == 0 )
	{
		return( -1 );
	}

	/* Compare The Strings
	---------------------- */
	while( --nLineLength )
	{
		/* Break When The End Of Either String Is Reached
		------------------------------------------------- */
		if( *s == 0 || *t == 0 )
		{
			break;
		}

		/* Otherwise, Compare The Next Character Pairs
		---------------------------------------------- */
		if( *s == *t || tolower( *s ) == *t || *s == tolower( *t ) )
		{
			++s;
			++t;
			continue;
		}
	}

	return( *s - *t );
}

// -------------------------------------------------------------------
// Extract a 'C' string entry from an INI file.

char *
INIGetString(
	char *pFileData,				// File Data Pointer
	UINT	nFileLength,			// Size Of Data At pFileData
	char *MainKey,					// INI Section Keyword
	char *MinorKey,				// INI Entry Keyword
	char *DefaultAns
	)
{
#ifdef DEBUG
	static char *PROCNAME = "INIGetString";
#endif
	static char Reply[ 256 ];	// Probably Need Smarter Storage...
	UINT	nLineLength;
	char	*pLineData, *s;

	/* Sanity Checks On Arguments
	----------------------------- */
	Assert( pFileData );
	Assert( nFileLength > 0 );
	Assert( MainKey && strlen( MainKey ) > 0 );
	Assert( MinorKey && strlen( MinorKey ) > 0 );

	if( !pFileData || nFileLength == 0
		|| strlen( MainKey ) == 0 || strlen( MinorKey ) == 0 )
	{
		return( DefaultAns );
	}

	/* Walk Each Line Of The INI File
	--------------------------------- */
	while( INIGetLineLength( pFileData, nFileLength, &nLineLength ) )
	{
		pLineData = pFileData;

		pFileData += nLineLength;
		nFileLength -= nLineLength;

		/* Strip Leading White
		---------------------- */
		while( nLineLength && isspace( *pLineData ) )
		{
			++pLineData;
			--nLineLength;
		}

		/* Look For '[' Section Keyword Delimiter
		----------------------------------------- */
		if( nLineLength > 0 )
		{
			if( *pLineData == '[' )
			{
				/* Found An INI Section Delimiter
				--------------------------------- */
				++pLineData;				// Skip Leading '['
				--nLineLength;

				/* Strip Leading White Before Section Keyword
				--------------------------------------------- */
				while( nLineLength && isspace( *pLineData ) )
				{
					++pLineData;
					--nLineLength;
				}

				if( nLineLength > 0 )
				{
					/* Found A Section Keyword
					-------------------------- */
					if( INIstrnicmp( pLineData, MainKey, strlen( MainKey ) ) == 0 )
					{
						/* Found Desired Section
						------------------------ */
						DEBUGTRACE( "INIGetString: Found INI Section\n" );

						/* Walk Each Line Of The Section
						-------------------------------- */
						while( INIGetLineLength( pFileData, nFileLength, &nLineLength ) )
						{
							pLineData = pFileData;

							pFileData += nLineLength;
							nFileLength -= nLineLength;

							/* Strip Leading White
							---------------------- */
							while( nLineLength && isspace( *pLineData ) )
							{
								++pLineData;
								--nLineLength;
							}

							if( nLineLength > 0 )
							{
								if( *pLineData == '[' )
								{
									/* Found The Next INI Section
									-----------------------------
								 	* Quit and return the default. The desired INI
								 	* section was found, but the desired entry
								 	* wasn't.
								 	*/
									break;
								}
								else if( INIstrnicmp( pLineData, MinorKey, strlen( MinorKey ) ) == 0 )
								{
									/* Have Found The Entry Line!!!
									------------------------------- */
									DEBUGTRACE( "INIGetString: Found INI Entry\n" );

									/* Now Find The String
									---------------------- */
									pLineData += strlen( MinorKey );
									nLineLength -= strlen( MinorKey );

									/* Strip White
									--------------
									 * Strip white (if any) between MinorKey and
									 * the '=' assignment character.
									 */
									while( nLineLength && isspace( *pLineData ) )
									{
										++pLineData;
										--nLineLength;
									}

									/* Skip The '=' Assignment Character
									------------------------------------ */
									if( nLineLength && *pLineData == '=' )
									{
										++pLineData;
										--nLineLength;
									}

									/* Strip More White
									-------------------
									 * Strip white (if any) between the '=' assignment
									 * character and the string.
									 */
									while( nLineLength && isspace( *pLineData ) )
									{
										++pLineData;
										--nLineLength;
									}

									/* Strip Trailing White
									----------------------- */
									if( nLineLength - 1 > 0 )
									{
										s = &pLineData[ nLineLength - 1 ];

										while( nLineLength - 1 > 0
											&& isspace( *s ) )
										{
											--s;
											--nLineLength;
										}
									}

									if( nLineLength )
									{
										/* Finally Found It!!!
										---------------------- */
										if( nLineLength < 255 )
										{
											memcpy( Reply, pLineData, nLineLength );
											Reply[ nLineLength ] = 0;

											return( Reply );
										}
									}
								
									DEBUGWARN( "INIGetString: Problem Handling MinorKey\n" );
								}
							}
						}

						DEBUGWARN( "INIGetString: MinorKey Not Found\n" );

						return( DefaultAns );
					}
				}
			}
		}
	}

	DEBUGWARN( "INIGetString: MainKey Not Found" );

	return( DefaultAns );
}


// -------------------------------------------------------------------
// For NDIS 3.0, use this function to read the protocol's adapter
// BINDING information from the PROTOCOL.INI file.
//
// For NDIS 3.1, use the NdisReadBinding wrapper function.

VOID
SNTRANSReadBinding(
    OUT PNDIS_STATUS Status,
    OUT PNDIS_STRING * Binding,
    IN PUNICODE_STRING RegistryPath		// NULL for VxDs
	)
{
#ifdef DEBUG
	static char *PROCNAME = "SNTRANSReadBinding";
#endif
	NDIS_HANDLE hFileHandle;
	UINT			nFileLength;
	char			*pFileData, *pINIString;
	NDIS_STRING szFileName = NDIS_STRING_CONST( "PROTOCOL.INI" );
	static NDIS_STRING AdapterName = NDIS_STRING_CONST( "Bananna" );

	/* Sanity Check On Arguments
	---------------------------- */
	Assert( Status );
	Assert( Binding );

	*Status = NDIS_STATUS_FAILURE;

	/* Open The NDIS 3.0 PROTOCOL.INI File
	-------------------------------------- */
	NdisOpenFile(
		Status,
		&hFileHandle,
		&nFileLength,
		&szFileName,
		-1								// No restriction on address...
		);

	if( *Status == NDIS_STATUS_SUCCESS )
	{
		/* Sanity Check On Returned Values
		---------------------------------- */
		Assert( hFileHandle );
		Assert( nFileLength );

		if( hFileHandle && nFileLength )
		{
			NdisMapFile(
				Status,
				&pFileData,
				hFileHandle
				);

			if( *Status == NDIS_STATUS_SUCCESS )
			{
				/* Sanity Check On Returned Values
				---------------------------------- */
				Assert( pFileData );

				pINIString = INIGetString(
									pFileData,
									nFileLength,
//									ATKN30_ID_STRING,
									TRANSPORT_NAME,
									"BINDINGS",
									NULL
									);

				if( pINIString )
				{
					/* BINDINGS String Entry Was Found
					---------------------------------- */
					NdisInitializeString( &AdapterName, pINIString );

					*Binding = &AdapterName;

					*Status == NDIS_STATUS_SUCCESS;
				}
				else
				{
					/* BINDINGS String Entry Was NOT Found
					-------------------------------------- */
					*Status = NDIS_STATUS_FAILURE;
				}
				 
				NdisUnmapFile( hFileHandle );
			}
		}

		NdisCloseFile( hFileHandle );
	}

	return;
}

// -------------------------------------------------------------------
// Handler for Control Message DEVICE_INIT (NDIS 3.0/WFW 3.11)
//
// NDIS 3.0 protocol drivers are statically loaded based on a value
// in the SYSTEM.INI file. Here is an example:
//
//     [386Enh]
//     ...
//     ; NOTE: secondnet=is synonomous with device=
//     secondnet=C:\WINDOWS\SYSTEM\SNTRANS.386
//
// NDIS 3.0 protocol drivers under Windows for Workgroups 3.11 must
// read the PROTOCOL.INI file to determing their binding, and then
// call NdisOpenAdapter themselves. This is actually what is done
// in SNTRANS. However, it is done in such a way that the commonality
// between NDIS 3.0 and NDIS 3.11 protocol drivers is maximized.
//
// In the SNTRANS implementation, OnDeviceInit() emulates the NDIS 3.1
// behavior. Specifically, OnDeviceInit() reads the PROTOCOL.INI file
// and calls the BindAdapterHandler with the binding information.
//
// Here are example PROTOCOL.INI entries for SNTRANS when bound to a
// NE2000 Ethernet driver:
//
//     [SNTRANS]
//     DriverName=SNTRANS
//     BINDINGS=MS$NE2000
//
// So, in the SNTRANS example, SNTRANSBindAdapterHandler is called for
// both NDIS 3.0/WFW 3.11 and NDIS 3.1/Win 95.
//

BOOL OnDeviceInit(VMHANDLE hVM, PCHAR CommandTail)
{
	NDIS_STATUS		nNdisStatus, nNdisStatus2;
	PNDIS_STRING	AdapterName;

	SETDEBUGLEVEL( DBG_ERROR, DBG_TRACE );

	DEBUGTRACE( "OnDeviceInit Entry..." );

	/* Read Binding Information From PROTOCOL.INI
	--------------------------------------------- */
	DEBUGTRACE( "Calling SNTRANSReadBinding..." );

	SNTRANSReadBinding(
		&nNdisStatus,
		&AdapterName,
		NULL
		);

	if( nNdisStatus != NDIS_STATUS_SUCCESS )
	{
		//
		// SNTRANSReadBinding failed.
		//

#ifdef DEBUG
		dprintf( "SNTRANSReadBinding failed with code 0x%X\n", nNdisStatus );
#endif

		return( FALSE );
	}

	/* Call The Common NDIS 3.0/3.1 Driver Entry Point
	-------------------------------------------------- */
	nNdisStatus = DriverEntry(NULL, NULL);
	
	if( nNdisStatus != NDIS_STATUS_SUCCESS)
	{
		return( FALSE );
	}

	DEBUGTRACE( "SNTRANSReadBinding succeeded\n" );

	DEBUGTRACE( "Calling SNTRANSBindAdapterHandler..." );

	/* Call Protocol To Bind To The Adapter
	---------------------------------------
	 * Note: This mechanism will need to be modified if the protocol
	 * binds to multiple adapters...
	 */
	SNTRANSBindAdapterHandler(
		&nNdisStatus,
		(NDIS_HANDLE )&SNTRANSContext.hNdisProtocolHandle,
		AdapterName,
		NULL,
		NULL
		);

	DEBUGTRACE( "SNTRANSBindAdapterHandler complete\n" );

	if( nNdisStatus != NDIS_STATUS_SUCCESS )
	{
		//
		// SNTRANSBindAdapterHandler failed.
		//

#ifdef DEBUG
		dprintf( "SNTRANSBindAdapterHandler failed with code 0x%X\n", nNdisStatus );
#endif

		SNTRANSUnloadProtocolHandler();

		return( FALSE );
	}

	return( TRUE );
}

#endif // WIN40
